//Project: roverControl
//Homepage: www.HomoFaciens.de
//Author Norbert Heinz
//Version: 2.0
//Creation date: 05.06.2011
//Copyright: (c) 2011 by Norbert Heinz
//This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
//This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with this program if not, see http://www.gnu.org/licenses/
//
//Program code based on PowerSwitch (author Christian Starkjohann) - Thanks!
//Program uses generic shared VID/PID pair of OBJECTIVE DEVELOPMENT Software GmbH - see documentation for licence and details

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <usb.h>    /* this is libusb, see http://libusb.sourceforge.net/ */

#define USBDEV_SHARED_VENDOR    0x16C0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   0x05DC  /* Obdev's free shared PID */
/* Use obdev's generic shared VID/PID pair and follow the rules outlined
 * in firmware/usbdrv/USBID-License.txt.
 */

#define RCCMD_ECHO           0
#define RCCMD_STEERING       1
#define RCCMD_FORWARD        2
#define RCCMD_BACKWARD       3
#define RCCMD_LIGHT          4
#define RCCMD_CAMERAHEIGHT   5
#define RCCMD_VERTICALANGLE  6
#define RCCMD_INIT          10
#define RCCMD_SLEEP         11
#define RCCMD_GETIRMODULE   12
#define RCCMD_WAITIRMODULE  13
#define RCCMD_SENDIRCOMMAND 14
/* These are the vendor specific SETUP commands implemented by our USB device */

static void usage(char *name)
{
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "  %s test\n", name);
    fprintf(stderr, "  %s f (forward) <duration>\n", name);
    fprintf(stderr, "  %s b (backward) <duration>\n", name);
    fprintf(stderr, "  %s s <steeringAngle>\n", name);
    fprintf(stderr, "  %s h <cameraHeight>\n", name);
    fprintf(stderr, "  %s v <vertikalAngle>\n", name);
    fprintf(stderr, "  %s l <light>\n", name);
    fprintf(stderr, "  %s c <m>\n", name);
    fprintf(stderr, "  %s o \n", name);
    fprintf(stderr, "  %s r (ask for infra red ID in range)\n", name);
    fprintf(stderr, "  %s c <commandNr>(send command via infra red interface)\n\n", name);

    fprintf(stderr, "duration is a value in the range 1...255\n");
    fprintf(stderr, "steeringAngle is a value in the range 60...90\n");
    fprintf(stderr, "verticalAngle is a value in the range 33...120\n");
    fprintf(stderr, "cameraHeight is a value in the range 83...100\n");
    fprintf(stderr, "light is a value in the range 0...200\n");
    fprintf(stderr, "commandNr is a value in the range 1...255\n");
}


static int  usbGetStringAscii(usb_dev_handle *dev, int index, int langid, char *buf, int buflen)
{
char    buffer[256];
int     rval, i;

    if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, buffer, sizeof(buffer), 100)) < 0)
        return rval;
    if(buffer[1] != USB_DT_STRING)
        return 0;
    if((unsigned char)buffer[0] < rval)
        rval = (unsigned char)buffer[0];
    rval /= 2;
    /* lossy conversion to ISO Latin1 */
    for(i=1;i<rval;i++){
        if(i > buflen)  /* destination buffer overflow */
            break;
        buf[i-1] = buffer[2 * i];
        if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
            buf[i-1] = '?';
    }
    buf[i-1] = 0;
    return i-1;
}

#define USB_ERROR_NOTFOUND  1
#define USB_ERROR_ACCESS    2
#define USB_ERROR_IO        3

static int usbOpenDevice(usb_dev_handle **device, int vendor, char *vendorName, int product, char *productName)
{
struct usb_bus      *bus;
struct usb_device   *dev;
usb_dev_handle      *handle = NULL;
int                 errorCode = USB_ERROR_NOTFOUND;
static int          didUsbInit = 0;

    if(!didUsbInit){
        didUsbInit = 1;
        usb_init();
    }
    usb_find_busses();
    usb_find_devices();
    for(bus=usb_get_busses(); bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next){
            if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product){
                printf ("This is our device: Vendor=0x%x, Product=0x%x\n", dev->descriptor.idVendor, dev->descriptor.idProduct);
                char    string[256];
                int     len;
                handle = usb_open(dev); /* we need to open the device in order to query strings */
                if(!handle){
                    errorCode = USB_ERROR_ACCESS;
                    fprintf(stderr, "Warning: cannot open USB device: %s\n", usb_strerror());
                    continue;
                }
                if(vendorName == NULL && productName == NULL){  /* name does not matter */
                    break;
                }
                /* now check whether the names match: */
                len = usbGetStringAscii(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
                if(len < 0){
                    errorCode = USB_ERROR_IO;
                    fprintf(stderr, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
                }
                else{
                    errorCode = USB_ERROR_NOTFOUND;
                    if(strcmp(string, vendorName) == 0){
                        len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
                        if(len < 0){
                            errorCode = USB_ERROR_IO;
                            fprintf(stderr, "Warning: cannot query product for device: %s\n", usb_strerror());
                        }else{
                            errorCode = USB_ERROR_NOTFOUND;
                            if(strcmp(string, productName) == 0)
                                break;
                        }
                    }
                }
                usb_close(handle);
                handle = NULL;
            }
            else{
              printf ("Not our device: Vendor=0x%x, Product=0x%x\n", dev->descriptor.idVendor, dev->descriptor.idProduct);
            }
        }
        if(handle)
            break;
    }
    if(handle != NULL){
        errorCode = 0;
        *device = handle;
    }
    return errorCode;
}


int main(int argc, char **argv)
{
usb_dev_handle      *handle = NULL;
unsigned char       buffer[8];
int                 nBytes;
int                 argsError = 0;
int                 steeringAngle = 0;
int                 duration = 0;
int                 cameraHeight = 30;
int                 verticalAngle = 30;
int                 light = 99;
int                 receiveByte = 0;
int                 commandNr = 0;
long                startTime = 0;
int                 oldIrId = 0;
FILE                *IRID;
char                IrFileName[]="IrModuleInRange.txt";
char                charTemp[1000]="";

    if(argc < 2){
        usage(argv[0]);
        exit(1);
    }

    if(strcmp(argv[1], "s") == 0){
      steeringAngle = atoi(argv[2]);
      if(steeringAngle != 0){
        if(steeringAngle < 60 || steeringAngle > 90){
          argsError = 1;
        }
      }
    }
    else if(strcmp(argv[1], "f") == 0){
      duration = atoi(argv[2]);
      if(duration < 1 || duration > 255){
        argsError = 1;
      }
    }
    else if(strcmp(argv[1], "b") == 0){
      duration = atoi(argv[2]);
      if(duration < 1 || duration > 255){
        argsError = 1;
      }
    }
    else if(strcmp(argv[1], "h") == 0){
      cameraHeight = atoi(argv[2]);
      if(cameraHeight != 0){
        if(cameraHeight < 83 || cameraHeight > 100){
          argsError = 1;
        }
      }
    }
    else if(strcmp(argv[1], "v") == 0){
      verticalAngle = atoi(argv[2]);
      if(verticalAngle != 0){
        if(verticalAngle < 33 || verticalAngle > 120){
          argsError = 1;
        }
      }
    }
    else if(strcmp(argv[1], "l") == 0){
      light = atoi(argv[2]);
      if(light < 0 || light > 200){
        argsError = 1;
      }
    }
    else if(strcmp(argv[1], "c") == 0){
      commandNr = atoi(argv[2]);
      if(commandNr < 1 || commandNr > 255){
        argsError = 1;
      }
    }

    if(argsError == 1){
      usage(argv[0]);
      exit(1);
    }

    usb_init();
    if(usbOpenDevice(&handle, USBDEV_SHARED_VENDOR, "www.HomoFaciens.de", USBDEV_SHARED_PRODUCT, "RoverControl") != 0){
        fprintf(stderr, "Could not find USB device \"RoverControl\" with vid=0x%x pid=0x%x\n", USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
        exit(1);
    }
/* We have searched all devices on all busses for our USB device above. Now
 * try to open it and perform the vendor specific control operations for the
 * function requested by the user.
 */
    if(strcmp(argv[1], "test") == 0){
        int i, v, r;
/* The test consists of writing 1000 random numbers to the device and checking
 * the echo. This should discover systematic bit errors (e.g. in bit stuffing).
 */
        for(i=0;i<1000;i++){
            v = rand() & 0xffff;
            nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_ECHO, v, 0, (char *)buffer, sizeof(buffer), 5000);
            if(nBytes < 2){
                if(nBytes < 0)
                    fprintf(stderr, "USB error: %s\n", usb_strerror());
                fprintf(stderr, "only %d bytes received in iteration %d\n", nBytes, i);
                fprintf(stderr, "value sent = 0x%x\n", v);
                exit(1);
            }
            r = buffer[0] | (buffer[1] << 8);
            if(r != v){
                fprintf(stderr, "data error: received 0x%x instead of 0x%x in iteration %d\n", r, v, i);
                exit(1);
            }
        }
        printf("test succeeded\n");
    }
    else if(strcmp(argv[1], "s") == 0){
      printf ("Adjusting servo position to %d.\n", steeringAngle);
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_STEERING, steeringAngle, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "f") == 0){
      printf ("Forward direction set, duration=%d.\n", duration);
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_FORWARD, duration, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "b") == 0){
      printf ("Reverse direction set, duration=%d.\n", duration);
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_BACKWARD, duration, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "h") == 0){
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_CAMERAHEIGHT, cameraHeight, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "v") == 0){
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_VERTICALANGLE, verticalAngle, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "l") == 0){
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_LIGHT, light, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "o") == 0){
      int signal = 'o';
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_INIT, signal, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "O") == 0){
      int signal = 'o';
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_SLEEP, signal, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
    }
    else if(strcmp(argv[1], "r") == 0){
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_GETIRMODULE, receiveByte, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 2){
        if(nBytes < 0)
        fprintf(stderr, "USB error: %s\n", usb_strerror());
        exit(1);
      }
      printf("IR module number=%d, SendStatus=%d\n", buffer[0], buffer[1]);
      oldIrId = buffer[0];
      buffer[1] = 1;
      while(buffer[1] > 0){
        nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_WAITIRMODULE, receiveByte, 0, (char *)buffer, sizeof(buffer), 100);
        if(nBytes < 2){
          if(nBytes < 0)
          fprintf(stderr, "USB error: %s\n", usb_strerror());
          printf("USB error: %s\n", usb_strerror());
          exit(1);
        }
        printf("IR module number=%d, SendStatus=%d\n", buffer[0], buffer[1]);
        startTime = time(0);
        while(time(0) - startTime < 1);
      }
      printf("IR module number=%d, SendStatus=%d\n", buffer[0], buffer[1]);
      if((IRID=fopen(IrFileName,"wb"))==NULL){
        printf("Can't open file '%s'!\n", IrFileName);
        return(1);
      }
      else{
        if(oldIrId == buffer[0]){
          printf("Equal!\n");
          sprintf(charTemp, "Module in range = %03d", buffer[0]);
          fwrite(&charTemp, 21, 1, IRID);
        }
        else{
          printf("NOT equal!\n");
          sprintf(charTemp, "Module in range = 000");
          fwrite(&charTemp, 21, 1, IRID);
        }
        fclose(IRID);
      }
    }
    else if(strcmp(argv[1], "c") == 0){
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_SENDIRCOMMAND, commandNr, 0, (char *)buffer, sizeof(buffer), 100);
      if(nBytes < 0){
          fprintf(stderr, "USB error: %s\n", usb_strerror());
      }
      while(buffer[1] > 0){
        nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, RCCMD_WAITIRMODULE, receiveByte, 0, (char *)buffer, sizeof(buffer), 100);
        if(nBytes < 2){
          if(nBytes < 0)
          fprintf(stderr, "USB error: %s\n", usb_strerror());
          printf("USB error: %s\n", usb_strerror());
          exit(1);
        }
        printf("IR module number=%d, SendStatus=%d\n", buffer[0], buffer[1]);
        startTime = time(0);
        while(time(0) - startTime < 1);
      }
      printf("commandNr=%d sent to module, nBytes=%d\n", commandNr, nBytes);
    }

    usb_close(handle);
    return 0;
}
